ACTIVATE_SCENARIO = "ACTIVATE_SCENARIO"
CALCULATION_DONE = "CALCULATION_DONE"
CALCULATION_FAILED = "CALCULATION_FAILED"
CASE_STUDY_ENDED = "CASE_STUDY_ENDED"
CASE_STUDY_INTERRUPT = "CASE_STUDY_INTERRUPT"
CASE_STUDY_LIST_CHANGED = "CASE_STUDY_LIST_CHANGED"
CASE_STUDY_NUMBER_CHANGED = "CASE_STUDY_NUMBER_CHANGED"
CASE_STUDY_PARAMETER_SELECTED = "CASE_STUDY_PARAMETER_SELECTED"
CASE_STUDY_PROGRESS = "CASE_STUDY_PROGRESS"
CASE_STUDY_PROPERTIES_SELECTED = "CASE_STUDY_PROPERTIES_SELECTED"
CASE_STUDY_RUN = "CASE_STUDY_RUN"
COPY_SCENARIO = "COPY_SCENARIO"
COPY_STREAM_TABLE = "COPY_STREAM_TABLE"
DELETE_SCENARIO = "DELETE_SCENARIO"
EXIT_APP = "EXIT_APP"
EXPORT_CANVAS_GFX = "EXPORT_CANVAS_GFX"
INITIALIZATION_DONE = "INITIALIZATION_DONE"
NEW_UNIT_DEFINED = "NEW_UNIT_DEFINED"
OPEN_FILE = "OPEN_FILE"
OPEN_RESULTS = "OPEN_RESULTS"
OPEN_SCENARIOS = "OPEN_SCENARIOS"
RENAME_SCENARIO = "RENAME_SCENARIO"
RESULT_UNIT_CHANGED = "RESULT_UNIT_CHANGED"
RESULT_UNIT_CLICKED = "RESULT_UNIT_CLICKED"
RUN_CASE_STUDY = "RUN_CASE_STUDY"
RUN_MODEL = "RUN_MODEL"
SAFE_FILE = "SAFE_FILE"
SAFE_FILE_AS = "SAFE_FILE_AS"
SHOW_PARAMETER_IN_CANVAS = "SHOW_PARAMETER_IN_CANVAS"
